import torch
import numpy as np
import torch.nn as nn
import torch.nn.functional as F
import matplotlib.pyplot as plt
import matplotlib
from matplotlib.colors import LinearSegmentedColormap
from utilities3 import *
from timeit import default_timer
from scipy import interpolate
from similaritymeasures import frechet_dist

def frechet(C1,C2,closed=True):
    if closed:
        n1 = len(C1)
        n2 = len(C2)
        
        if n1>n2:
            n1,n2 = n2,n1
            C1,C2 = C2,C1
            
        C2 = np.vstack((C2,C2[0]))
        i = np.arange(n1)
        q = np.inf
        for j in range(n1):
            ids = np.hstack((np.roll(i,j),[-j]))
            q = np.min([q,frechet_dist(C1[ids],C2)])
            
        return q
    else:
        return frechet_dist(C1,C2)

def calculate_area(x, y):
    n = len(x)
    area = 0

    for i in range(n):
        j = (i + 1) % n
        area += (x[i] * y[j]) - (x[j] * y[i])

    return abs(area) / 2.0


def my_to_torch(x):
    if not torch.is_tensor(x):
        x = x.astype(np.float32)
        return torch.from_numpy(x)
    else:
        return x

def tikzplot(file,X,Y,style='',forget=True):
   with open(file,"w") as fid:
       if forget:
           style = style+',forget plot'

       if len(style)>0:
           fid.write('\\addplot['+style+']\n')
       else:
           fid.write('\\addplot\n')
       fid.write('table[row sep=crcr]{\n')
       dataXY = [X,Y]
       for x in zip(*dataXY):
           fid.write("%1.8f\t%1.8f\\\\\n" % (x))
       fid.write('};\n')

################################################################
# fourier layer
################################################################

class SpectralConv1d(nn.Module):
    def __init__(self, in_channels, out_channels, modes1):
        super(SpectralConv1d, self).__init__()

        """
        1D Fourier layer. It does FFT, linear transform, and Inverse FFT.    
        """

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.modes1 = modes1  #Number of Fourier modes to multiply, at most floor(N/2) + 1

        self.scale = (1 / (in_channels*out_channels))
        # weights are initialised and are now considered as learnable parameters
        self.weights1 = nn.Parameter(self.scale * torch.rand(in_channels, out_channels, self.modes1, dtype=torch.cfloat))

    # Complex multiplication
    def compl_mul1d(self, input, weights):
        # (batch, in_channel, x ), (in_channel, out_channel, x) -> (batch, out_channel, x)
        return torch.einsum("bix,iox->box", input, weights)

    def forward(self, x):
        batchsize = x.shape[0]
        #Compute Fourier coeffcients up to factor of e^(- something constant)
        x_ft = torch.fft.rfft(x)

        # Multiply relevant Fourier modes
        out_ft = torch.zeros(batchsize, self.out_channels, x.size(-1)//2 + 1,  device=x.device, dtype=torch.cfloat)
        out_ft[:, :, :self.modes1] = self.compl_mul1d(x_ft[:, :, :self.modes1], self.weights1)

        #Return to physical space
        x = torch.fft.irfft(out_ft, n=x.size(-1))
        return x

class FNO1d(nn.Module):
    def __init__(self, modes, width):
        super(FNO1d, self).__init__()

        """
        The overall network. It contains 4 layers of the Fourier layer.
        1. Lift the input to the desire channel dimension by self.fc0 .
        2. 4 layers of the integral operators u' = (W + K)(u).
            W defined by self.w; K defined by self.conv .
        3. Project from the channel space to the output space by self.fc1 and self.fc2 .
        
        input: the solution of the first 10 timesteps and the grid (a(x), x)
        input shape: (batchsize, x=s, c=11)
        output: the solution of the next timestep
        output shape: (batchsize, x=s, c=1)
        """

        self.modes1 = modes
        self.width = width
        # Applies a linear transformation to the incoming data
        # nn.Linear(in_features, out_features, ...)
        # input channel is 41
        self.fc0 = nn.Linear(41, self.width)
        # Apply a 1D convolution in the spectral case (from class SpectralConv1d)
        self.conv0 = SpectralConv1d(self.width, self.width, self.modes1)
        self.conv1 = SpectralConv1d(self.width, self.width, self.modes1)
        # self.conv2 = SpectralConv1d(self.width, self.width, self.modes1)
        # self.conv3 = SpectralConv1d(self.width, self.width, self.modes1)
        # self.conv4 = SpectralConv1d(self.width, self.width, self.modes1)
        # self.conv5 = SpectralConv1d(self.width, self.width, self.modes1)
        # self.conv6 = SpectralConv1d(self.width, self.width, self.modes1)
        # self.conv7 = SpectralConv1d(self.width, self.width, self.modes1)
        # Apply a 1D convolution over an input signal
        # torch.nn.Conv1d(in_channels, out_channels, kernel_size,...)
        self.w0 = nn.Conv1d(self.width, self.width, 1)
        self.w1 = nn.Conv1d(self.width, self.width, 1)
        # self.w2 = nn.Conv1d(self.width, self.width, 1)
        # self.w3 = nn.Conv1d(self.width, self.width, 1)
        # self.w4 = nn.Conv1d(self.width, self.width, 1)
        # self.w5 = nn.Conv1d(self.width, self.width, 1)
        # self.w6 = nn.Conv1d(self.width, self.width, 1)
        # self.w7 = nn.Conv1d(self.width, self.width, 1)
        # This is the step before the last application of the non-linearity
        self.fc1 = nn.Linear(self.width, 128)
        # This is the final step, going from 128 to 3 channels (i.e. the solution u(x))
        self.fc2 = nn.Linear(128, 3)

    def forward(self, x):
        # first step, raise the input to a higher-dimension space
        # from 2 to self.width channels
        # i.e. from [batch_size,s,2] to [batch_size,s,self.width]
        x = self.fc0(x)
        # change the ordering of the columns of x
        # from [batch_size,s,self.width] to [batch_size,self.width,s]
        x = x.permute(0, 2, 1)
        
        # Fourier layer (1)
        x1 = self.conv0(x)
        x2 = self.w0(x)
        x = x1 + x2
        x = F.relu(x)

        # Fourier layer (2)
        x1 = self.conv1(x)
        x2 = self.w1(x)
        x = x1 + x2
        # x = F.relu(x)

        # # Fourier layer (3)
        # x1 = self.conv2(x)
        # x2 = self.w2(x)
        # x = x1 + x2
        # x = F.relu(x)

        # # Fourier layer (4)
        # x1 = self.conv3(x)
        # x2 = self.w3(x)
        # x = x1 + x2
        # # x = F.relu(x)
        
        # # Fourier layer (5)
        # x1 = self.conv4(x)
        # x2 = self.w4(x)
        # x = x1 + x2
        # x = F.relu(x)

        # # Fourier layer (6)
        # x1 = self.conv5(x)
        # x2 = self.w5(x)
        # x = x1 + x2
        # x = F.relu(x)

        # # Fourier layer (7)
        # x1 = self.conv6(x)
        # x2 = self.w6(x)
        # x = x1 + x2
        # x = F.relu(x)

        # # Fourier layer (8)
        # x1 = self.conv7(x)
        # x2 = self.w7(x)
        # x = x1 + x2

        # change the ordering of the columns of x
        # from [batch_size,self.width,s] to [batch_size,s,self.width]
        x = x.permute(0, 2, 1)
        # raise the solution even further, from [batch_size,s,self.width] to [batch_size,s,128]
        x = self.fc1(x)
        # apply the final non-linearity
        x = F.relu(x)
        # bring the solution back to [batch_size,s,1]
        x = self.fc2(x)
        return x
   
   

################################################################
# configs
################################################################
TEST_PATH  = '../../data/testing/eq10_specific.npy'
HET_PATH   = '../../data/testing/eq10_specific-het.npy'
model_name = 'model1600_m8_w128'

params = np.load('model/'+model_name+'_other_params.npy')
model  = FNO1d(params[0],params[1])

checkpoint = torch.load('model/'+model_name)
model.load_state_dict(checkpoint['model_state_dict'])
model.eval()

print(count_params(model))

sub   = 1
T_in  = 10
T     = 90
step  = 1
batch_size = 1

################################################################
# load data
################################################################
L = 15
def het(d,x,y):
     x = np.cos(d[6])*(x/L) - np.sin(d[6])*(y/L)
     y = np.sin(d[6])*(x/L) + np.cos(d[6])*(y/L)
     return 1 + d[0]*x + d[1]*np.tanh(d[2]*np.cos(d[3]*np.pi*(x*np.sin(d[5])+np.cos(d[5])*y))*\
                                           np.cos(d[4]*np.pi*x))
#
index      = 89 # Which time instance to plot?
samples    = 6
batch_size = 1
data          = np.load(TEST_PATH)
nan_locations = np.argwhere(np.isnan(data))
values        = np.unique(nan_locations[:,0])
data          = np.delete(data,values,axis=0)
S             = data.shape[3]
#
u    = np.arange(S) * 2*np.pi/S
grid = u.reshape(1, S, 1)
grid = torch.tensor(grid, dtype=torch.float)

Xmin=0; Xmax=L
Ymin=0; Ymax=L
xf = np.linspace(Xmin,Xmax,1000)
yf = np.linspace(Ymin,Ymax,1000)
xff,yff = np.meshgrid(xf,yf)

for sample in range(samples):

    data_het   = np.delete(np.load(HET_PATH),values,axis=0)[sample]
       
    #
    heter     = torch.empty(1,S,T_in)
    heter2    = torch.empty(batch_size,S,1)
    
    x_cl = my_to_torch(data[sample,0,:T_in,:]).reshape(1,T_in,S).permute(0,2,1)
    y_cl = my_to_torch(data[sample,1,:T_in,:]).reshape(1,T_in,S).permute(0,2,1)
    
    heter  = my_to_torch(het(data_het,x_cl[0],y_cl[0]))
    #
    test_a     = my_to_torch(data[sample,:,:T_in      ,::sub]).reshape(1,data.shape[1],T_in,S).permute(0,1,3,2)
    centroid_x = torch.mean(test_a[:,0,:,:],1).reshape(1,1,T_in).repeat(1,S,1)
    centroid_y = torch.mean(test_a[:,1,:,:],1).reshape(1,1,T_in).repeat(1,S,1)
    rx         = test_a[:,0,:,:T_in]-centroid_x
    ry         = test_a[:,1,:,:T_in]-centroid_y
    r          = (rx**2+ry**2)**0.5
    test_a     = torch.cat([centroid_x,centroid_y,r], dim=-1)
    test_a     = torch.cat([test_a, heter.reshape(1,S,T_in)], dim=-1)
    test_a     = torch.cat([test_a, grid], dim=-1)
    #
    test_u = my_to_torch(data[sample,:,T_in:T+T_in,::sub]).reshape(1,data.shape[1],T,S).permute(0,1,3,2)
    centroid_x = torch.mean(test_u[:,0,:,:],1).reshape(1,1,T).repeat(1,S,1)
    centroid_y = torch.mean(test_u[:,1,:,:],1).reshape(1,1,T).repeat(1,S,1)
    rx         = test_u[:,0,:,:T]-centroid_x
    ry         = test_u[:,1,:,:T]-centroid_y
    r          = (rx**2+ry**2)**0.5
    test_u     = torch.cat([centroid_x,centroid_y,r], dim=-1)
    
    ################################################################
    # evaluation
    ################################################################
    
    #print(model.state_dict(), file=open('output.txt', 'a'))
    t1 = default_timer()
    #
    pred = torch.zeros(test_u.shape)
    myloss    = LpLoss(size_average=False)
    xx = test_a
    yy = test_u
    loss = 0
    test_l2_step = 0
    for t in range(0, T, step):
        # 
        y = torch.cat([yy[..., t:t + step],yy[...,T+t:T+t+step],yy[...,2*T+t:2*T+t+step]],-1)
        #
        im = model(xx)
        #
        x_loc     = im[..., 0]  +im[...,2]*torch.cos(grid[0,:,0])
        y_loc     = im[..., 1]  +im[...,2]*torch.sin(grid[0,:,0])
        x_loc_ref = yy[..., t:t + step]  +yy[...,2*T+t:2*T+t+step]*torch.cos(grid)
        y_loc_ref = yy[..., T+t:T+t+step]+yy[...,2*T+t:2*T+t+step]*torch.sin(grid)
        
        loss = myloss(torch.cat([x_loc,y_loc]).reshape(batch_size, -1), torch.cat([x_loc_ref,y_loc_ref]).reshape(batch_size, -1))
        if t == 0:
            pred = im
        else:
            pred = torch.cat((pred, im), -1)
        #
        xx_i = torch.cat([xx[..., step:T_in],               \
                          torch.mean(im[...,0],1).reshape(batch_size,1,1).repeat(1,S,1),\
                          xx[..., T_in+step:2*T_in],        \
                          torch.mean(im[...,1],1).reshape(batch_size,1,1).repeat(1,S,1),\
                          xx[...,2*T_in+step:3*T_in],       \
                          im[...,2].reshape(batch_size,S,1) \
                          ], dim=-1).to(device)
        #
        x_cl = torch.mean(im[...,0],1).reshape(batch_size,1,1).repeat(1,S,1)+im[...,2].reshape(batch_size,S,1)*np.cos(grid.repeat(batch_size,1,1))
        y_cl = torch.mean(im[...,1],1).reshape(batch_size,1,1).repeat(1,S,1)+im[...,2].reshape(batch_size,S,1)*np.sin(grid.repeat(batch_size,1,1))
        #
        heter2[0,:,0] = my_to_torch(het(data_het, \
                                        x_cl[0,:,0].cpu().detach().numpy(), \
                                        y_cl[0,:,0].cpu().detach().numpy()))   
        #
        xx = torch.cat([xx_i.cpu(),                               \
                        xx[...,3*T_in+step:4*T_in],         \
                        heter2.reshape(batch_size,S,1),     \
                        grid.repeat(batch_size,1,1),        \
                        ], dim=-1    )
    
        test_l2_step += loss.item()
           
    #
    t2 = default_timer()
    print(f'prediction wall time = {t2-t1:.3f} s')
    x_loc = x_loc[0].detach().numpy(); x_loc_ref = x_loc_ref[0,:,0].detach().numpy();
    y_loc = y_loc[0].detach().numpy(); y_loc_ref = y_loc_ref[0,:,0].detach().numpy();
    r_ref = yy[...,2*T+t:2*T+t+step].detach().numpy()
    
       
    ########################
    # plotting heterogeneity
    ########################
    het_array = het(data_het,xff,yff)
    cmap = LinearSegmentedColormap.from_list('custom greys', [(1,1,1),(0.6,0.6,0.6)], N=256)
    matplotlib.image.imsave("figures/het-"+str(sample)+".png",het_array,vmin=1,vmax=2,cmap=cmap,origin='lower')
    ########################
    # plotting solution
    ########################
    size_title  = 20
    size_labels = 18
    size_ticks  = 15
    size_legend = 12
    
    fig, ax = plt.subplots(figsize=(5,5))
    
    ax.imshow(het_array,extent=[0, L, 0, L],cmap=cmap,origin='lower')
    
    out_x      = pred[0,:,index*3  ].detach()+pred[0,:,index*3+2]*torch.cos(grid[0,:,0])
    out_y      = pred[0,:,index*3+1].detach()+pred[0,:,index*3+2]*torch.sin(grid[0,:,0])
    out_true_x = test_u[0,:,index  ].detach()+test_u[0,:,2*T+index]*torch.cos(grid[0,:,0])
    out_true_y = test_u[0,:,T+index].detach()+test_u[0,:,2*T+index]*torch.sin(grid[0,:,0])
    out_true_x0 = data[sample,0,0,::sub]
    out_true_y0 = data[sample,1,0,::sub]
    # ODE
    plt.fill(out_true_x,out_true_y,facecolor=(0.45,0.7,1,0.5),edgecolor='tab:blue',lw=1.5)
    plt.plot([],[],color='tab:blue',lw=1.5)
    # FNO
    plt.fill(out_x.detach(),out_y.detach(),fill=False,edgecolor='tab:orange',lw=1.5)
    plt.plot([],[],color='tab:orange',lw=1.5)
    # initial
    plt.fill(out_true_x0,out_true_y0,facecolor=(0.45,0.7,1,0.5),edgecolor='tab:orange',lw=1.5)
    #
    plt.xlim([0, L])
    plt.ylim([0, L])
    plt.xticks(fontsize = size_ticks)
    plt.yticks(fontsize = size_ticks)
    plt.savefig('figures/'+str(sample)+'.png',bbox_inches='tight',transparent=True,format='png',dpi=300)
    
    out_x = out_x.detach().numpy()
    out_y = out_y.detach().numpy()
    out_true_x = out_true_x.detach().numpy()
    out_true_y = out_true_y.detach().numpy()
    tikzplot('figures/'+str(sample)+'_FNO.txt',np.append(out_x,out_x[0]),np.append(out_y,out_y[0]),style='pyorange,thick')
    tikzplot('figures/'+str(sample)+'_ref.txt',np.append(out_true_x,out_true_x[0]),np.append(out_true_y,out_true_y[0]),style='pyblue,thick,fill=pyblue,fill opacity=0.5')
    tikzplot('figures/'+str(sample)+'_ini.txt',np.append(out_true_x0,out_true_x0[0]),np.append(out_true_y0,out_true_y0[0]),style='pyorange,thick,fill=pyblue,fill opacity=0.5')


    Area  = calculate_area(out_true_x,out_true_y) 
    error = frechet(np.array([out_x,out_y]).T,np.array([out_true_x,out_true_y]).T)*np.sqrt(np.pi/Area)
    print(f'auxiliary error = {error:.3%}')
