
import torch
import numpy as np
import torch.nn as nn
import torch.nn.functional as F
from   scipy import interpolate
# import matplotlib as mpl
# mpl.use('Agg') # use matplotlib without graphics
import matplotlib.pyplot as plt
from utilities3 import *

import operator
from functools import reduce
from functools import partial

from timeit import default_timer

from Adam import Adam

import sys

class Unbuffered(object):
   def __init__(self, stream):
       self.stream = stream
   def write(self, data):
       self.stream.write(data)
       self.stream.flush()
   def writelines(self, datas):
       self.stream.writelines(datas)
       self.stream.flush()
   def __getattr__(self, attr):
       return getattr(self.stream, attr)


def my_to_torch(x):
    if not torch.is_tensor(x):
        x = x.astype(np.float32)
        return torch.from_numpy(x)
    else:
        return x

################################################################
# fourier layer
################################################################

class SpectralConv1d(nn.Module):
    def __init__(self, in_channels, out_channels, modes1):
        super(SpectralConv1d, self).__init__()

        """
        1D Fourier layer. It does FFT, linear transform, and Inverse FFT.    
        """

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.modes1 = modes1  #Number of Fourier modes to multiply, at most floor(N/2) + 1

        self.scale = (1 / (in_channels*out_channels))
        # weights are initialised and are now considered as learnable parameters
        self.weights1 = nn.Parameter(self.scale * torch.rand(in_channels, out_channels, self.modes1, dtype=torch.cfloat))

    # Complex multiplication
    def compl_mul1d(self, input, weights):
        # (batch, in_channel, x ), (in_channel, out_channel, x) -> (batch, out_channel, x)
        return torch.einsum("bix,iox->box", input, weights)

    def forward(self, x):
        batchsize = x.shape[0]
        #Compute Fourier coeffcients up to factor of e^(- something constant)
        x_ft = torch.fft.rfft(x)

        # Multiply relevant Fourier modes
        out_ft = torch.zeros(batchsize, self.out_channels, x.size(-1)//2 + 1,  device=x.device, dtype=torch.cfloat)
        out_ft[:, :, :self.modes1] = self.compl_mul1d(x_ft[:, :, :self.modes1], self.weights1)

        #Return to physical space
        x = torch.fft.irfft(out_ft, n=x.size(-1))
        return x

class FNO1d(nn.Module):
    def __init__(self, modes, width):
        super(FNO1d, self).__init__()

        """
        The overall network. It contains 4 layers of the Fourier layer.
        1. Lift the input to the desire channel dimension by self.fc0 .
        2. 4 layers of the integral operators u' = (W + K)(u).
            W defined by self.w; K defined by self.conv .
        3. Project from the channel space to the output space by self.fc1 and self.fc2 .
        
        input: the solution of the first 10 timesteps and the grid (a(x), x)
        input shape: (batchsize, x=s, c=11)
        output: the solution of the next timestep
        output shape: (batchsize, x=s, c=1)
        """

        self.modes1 = modes
        self.width = width
        # Applies a linear transformation to the incoming data
        # nn.Linear(in_features, out_features, ...)
        # input channel is 41
        self.fc0 = nn.Linear(41, self.width)
        # Apply a 1D convolution in the spectral case (from class SpectralConv1d)
        self.conv0 = SpectralConv1d(self.width, self.width, self.modes1)
        self.conv1 = SpectralConv1d(self.width, self.width, self.modes1)
        #self.conv2 = SpectralConv1d(self.width, self.width, self.modes1)
        #self.conv3 = SpectralConv1d(self.width, self.width, self.modes1)
        # self.conv4 = SpectralConv1d(self.width, self.width, self.modes1)
        # self.conv5 = SpectralConv1d(self.width, self.width, self.modes1)
        # self.conv6 = SpectralConv1d(self.width, self.width, self.modes1)
        # self.conv7 = SpectralConv1d(self.width, self.width, self.modes1)
        # Apply a 1D convolution over an input signal
        # torch.nn.Conv1d(in_channels, out_channels, kernel_size,...)
        self.w0 = nn.Conv1d(self.width, self.width, 1)
        self.w1 = nn.Conv1d(self.width, self.width, 1)
        #self.w2 = nn.Conv1d(self.width, self.width, 1)
        #self.w3 = nn.Conv1d(self.width, self.width, 1)
        # self.w4 = nn.Conv1d(self.width, self.width, 1)
        # self.w5 = nn.Conv1d(self.width, self.width, 1)
        # self.w6 = nn.Conv1d(self.width, self.width, 1)
        # self.w7 = nn.Conv1d(self.width, self.width, 1)
        # This is the step before the last application of the non-linearity
        self.fc1 = nn.Linear(self.width, 128)
        # This is the final step, going from 128 to 3 channels (i.e. the solution u(x))
        self.fc2 = nn.Linear(128, 3)

    def forward(self, x):
        # first step, raise the input to a higher-dimension space
        # from 2 to self.width channels
        # i.e. from [batch_size,s,2] to [batch_size,s,self.width]
        x = self.fc0(x)
        # change the ordering of the columns of x
        # from [batch_size,s,self.width] to [batch_size,self.width,s]
        x = x.permute(0, 2, 1)
        
        # Fourier layer (1)
        x1 = self.conv0(x)
        x2 = self.w0(x)
        x = x1 + x2
        x = F.relu(x)

        # Fourier layer (2)
        x1 = self.conv1(x)
        x2 = self.w1(x)
        x = x1 + x2
       # x = F.relu(x)

       # # Fourier layer (3)
       # x1 = self.conv2(x)
       # x2 = self.w2(x)
       # x = x1 + x2
       # x = F.relu(x)

       # # Fourier layer (4)
       # x1 = self.conv3(x)
       # x2 = self.w3(x)
       # x = x1 + x2
        # x = F.relu(x)
        
        # # Fourier layer (5)
        # x1 = self.conv4(x)
        # x2 = self.w4(x)
        # x = x1 + x2
        # x = F.relu(x)

        # # Fourier layer (6)
        # x1 = self.conv5(x)
        # x2 = self.w5(x)
        # x = x1 + x2
        # x = F.relu(x)

        # # Fourier layer (7)
        # x1 = self.conv6(x)
        # x2 = self.w6(x)
        # x = x1 + x2
        # x = F.relu(x)

        # # Fourier layer (8)
        # x1 = self.conv7(x)
        # x2 = self.w7(x)
        # x = x1 + x2

        # change the ordering of the columns of x
        # from [batch_size,self.width,s] to [batch_size,s,self.width]
        x = x.permute(0, 2, 1)
        # raise the solution even further, from [batch_size,s,self.width] to [batch_size,s,128]
        x = self.fc1(x)
        # apply the final non-linearity
        x = F.relu(x)
        # bring the solution back to [batch_size,s,1]
        x = self.fc2(x)
        return x
   
   

################################################################
# configs
################################################################

TRAIN_PATH = '../../data/training/eq10.npy'
TEST_PATH  = '../../data/training/eq10.npy'
HET_PATH   = '../../data/training/eq10-het.npy'
#
torch.manual_seed(0)
np.random.seed(0)
#
ntrain     = 1600
ntest      = 400
modes      = 8
width      = 128
batch_size = 20
#
epochs          = 1000
learning_rate   = 1e-3
scheduler_step  = 100
scheduler_gamma = 0.5
#
# sys.stdout = open("log_n"+str(ntrain)+ '_m' + str(modes) + '_w' + str(width)+'.out', 'w')
# sys.stdout = Unbuffered(sys.stdout)
#
print('epochs           = '+str(epochs))
print('iterations       = '+str(int(epochs*ntrain/batch_size)))
print('learning rate    = '+str(learning_rate))
print('scheduler step   = '+str(scheduler_step))
print('scheduler gamma  = '+str(scheduler_gamma))
#
path           = 'model'+str(ntrain)+ '_m' + str(modes) + '_w' + str(width)
path_model     = 'model/'+path
path_plots     = 'figures/'+path
#
np.save(path_model+'_other_params',[modes,width])
#
sub  = 1   # sub-sampling
T_in = 10  # number of input times
T    = 90  
step = 1   # prediction step, should be 1
#
plot_error = np.array([[],[]])

################################################################
# load data
################################################################
data          = np.load(TRAIN_PATH)
nan_locations = np.argwhere(np.isnan(data))
values        = np.unique(nan_locations[:,0])
data          = np.delete(data,values,axis=0)[:ntrain+ntest]
S             = data.shape[3]
#
u    = np.arange(S) * 2*np.pi/S
grid = u.reshape(1, S, 1)
grid = torch.tensor(grid, dtype=torch.float)
#
# Surface wetting profile
data_het   = np.delete(np.load(HET_PATH),values,axis=0)[:ntrain+ntest]
L = 15
def het(d,x,y):
    x = np.cos(d[6])*(x/L) - np.sin(d[6])*(y/L)
    y = np.sin(d[6])*(x/L) + np.cos(d[6])*(y/L)
    return 1 + d[0]*x + d[1]*np.tanh(d[2]*np.cos(d[3]*np.pi*(x*np.sin(d[5])+np.cos(d[5])*y))*\
                                          np.cos(d[4]*np.pi*x))

#
heter     = torch.empty(ntrain+ntest,S,T_in)
heter2    = torch.empty(batch_size,S,1)

x_cl = my_to_torch(data[:,0,:T_in,:]).permute(0,2,1)
y_cl = my_to_torch(data[:,1,:T_in,:]).permute(0,2,1)

for i in range(ntrain+ntest):
    heter[i]  = my_to_torch(het(data_het[i],x_cl[i],y_cl[i]))
#
train_a    = my_to_torch(data[:ntrain,:,:T_in      ,::sub]).permute(0,1,3,2)
centroid_x = torch.mean(train_a[:,0,:,:],1).reshape(ntrain,1,T_in).repeat(1,S,1)
centroid_y = torch.mean(train_a[:,1,:,:],1).reshape(ntrain,1,T_in).repeat(1,S,1)
rx         = train_a[:,0,:,:T_in]-centroid_x
ry         = train_a[:,1,:,:T_in]-centroid_y
r          = (rx**2+ry**2)**0.5
train_a    = torch.cat([centroid_x,centroid_y,r], dim=-1)
train_a    = torch.cat([train_a, heter[:ntrain]], dim=-1)
train_a    = torch.cat([train_a, grid.repeat(ntrain,1,1)], dim=-1)
extra      = my_to_torch(np.array([i for i in range(ntrain)])).reshape(ntrain,1,1).repeat(1,S,1)
train_a    = torch.cat([train_a, extra], dim=-1)
#
train_u    = my_to_torch(data[:ntrain,:,T_in:T+T_in,::sub]).permute(0,1,3,2)
centroid_x = torch.mean(train_u[:,0,:,:],1).reshape(ntrain,1,T).repeat(1,S,1)
centroid_y = torch.mean(train_u[:,1,:,:],1).reshape(ntrain,1,T).repeat(1,S,1)
rx         = train_u[:,0,:,:T]-centroid_x
ry         = train_u[:,1,:,:T]-centroid_y
r          = (rx**2+ry**2)**0.5
train_u    = torch.cat([centroid_x,centroid_y,r], dim=-1)
#
test_a     = my_to_torch(data[-ntest:,:,:T_in      ,::sub]).permute(0,1,3,2)
centroid_x = torch.mean(test_a[:,0,:,:],1).reshape(ntest,1,T_in).repeat(1,S,1)
centroid_y = torch.mean(test_a[:,1,:,:],1).reshape(ntest,1,T_in).repeat(1,S,1)
rx         = test_a[:,0,:,:T_in]-centroid_x
ry         = test_a[:,1,:,:T_in]-centroid_y
r          = (rx**2+ry**2)**0.5
test_a     = torch.cat([centroid_x,centroid_y,r], dim=-1)
test_a     = torch.cat([test_a, heter[-ntest:]], dim=-1)
test_a     = torch.cat([test_a, grid.repeat(ntest,1,1)], dim=-1)
extra      = my_to_torch(np.array([i+ntrain for i in range(ntest)])).reshape(ntest,1,1).repeat(1,S,1)
test_a     = torch.cat([test_a, extra], dim=-1)
#
test_u = my_to_torch(data[-ntest:,:,T_in:T+T_in,::sub]).permute(0,1,3,2)
centroid_x = torch.mean(test_u[:,0,:,:],1).reshape(ntest,1,T).repeat(1,S,1)
centroid_y = torch.mean(test_u[:,1,:,:],1).reshape(ntest,1,T).repeat(1,S,1)
rx         = test_u[:,0,:,:T]-centroid_x
ry         = test_u[:,1,:,:T]-centroid_y
r          = (rx**2+ry**2)**0.5
test_u     = torch.cat([centroid_x,centroid_y,r], dim=-1)

train_loader = torch.utils.data.DataLoader(torch.utils.data.TensorDataset(train_a, train_u), batch_size=batch_size, shuffle=True )
test_loader  = torch.utils.data.DataLoader(torch.utils.data.TensorDataset(test_a ,  test_u), batch_size=batch_size, shuffle=False)



################################################################
# training and evaluation
################################################################

model = FNO1d(modes,width)#.cuda()
model.to(device)

print('model parameters = {}'.format(count_params(model)))

optimizer = Adam(model.parameters(), lr=learning_rate, weight_decay=1e-4)
scheduler = torch.optim.lr_scheduler.StepLR(optimizer, step_size=scheduler_step, gamma=scheduler_gamma)
min_error = 1.0
myloss    = LpLoss(size_average=False)

for ep in range(epochs):
    model.train()
    t1 = default_timer()
    train_l2_step = 0
    train_l2_full = 0
    for xx, yy in train_loader:
        loss = 0
        xx = xx.to(device)
        yy = yy.to(device)
        #
        for t in range(0, T, step):
            #
            y  = torch.cat([yy[..., t:t + step],yy[...,T+t:T+t+step],yy[...,2*T+t:2*T+t+step]],-1)
            #
            im = model(xx[...,:-1])
            #
            loss += myloss(im.reshape(batch_size, -1), y.reshape(batch_size, -1))
            #
            if t == 0:
                pred = im
            else:
                pred = torch.cat((pred, im), -1)
            #
            xx_i = torch.cat([xx[..., step:T_in],               \
                              torch.mean(im[...,0],1).reshape(batch_size,1,1).repeat(1,S,1),\
                              xx[..., T_in+step:2*T_in],        \
                              torch.mean(im[...,1],1).reshape(batch_size,1,1).repeat(1,S,1),\
                              xx[...,2*T_in+step:3*T_in],       \
                              im[...,2].reshape(batch_size,S,1) \
                             ], dim=-1).to(device)
            #
            x_cl = torch.mean(im[...,0],1).reshape(batch_size,1,1).repeat(1,S,1)+im[...,2].reshape(batch_size,S,1)*torch.cos(grid.repeat(batch_size,1,1)).to(device)
            y_cl = torch.mean(im[...,1],1).reshape(batch_size,1,1).repeat(1,S,1)+im[...,2].reshape(batch_size,S,1)*torch.sin(grid.repeat(batch_size,1,1)).to(device)
            #
            k = 0
            for i in xx[:,0,-1]:
                ii = int(i.cpu().detach().numpy())
                heter2[k,:,0] = my_to_torch(het(data_het[ii], \
                                                x_cl[k,:,0].cpu().detach().numpy(), \
                                                y_cl[k,:,0].cpu().detach().numpy()))   
                k += 1
            #
            xx = torch.cat([xx_i,                               \
                            xx[...,3*T_in+step:4*T_in],         \
                            heter2.reshape(batch_size,S,1).to(device),     \
                            grid.repeat(batch_size,1,1).to(device),        \
                            xx[...,-1].reshape(batch_size,S,1), \
                           ], dim=-1    ) 
            
        train_l2_step += loss.item()
        l2_full = myloss(torch.cat([pred[...,0::3],pred[...,1::3],pred[...,2::3]],-1).reshape(batch_size, -1), yy.reshape(batch_size, -1))
        train_l2_full += l2_full.item()

        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

    test_l2_step = 0
    test_l2_full = 0
    with torch.no_grad():
        for xx, yy in test_loader:
            loss = 0
            xx = xx.to(device)
            yy = yy.to(device)
            #
            for t in range(0, T, step):
                # 
                y = torch.cat([yy[..., t:t + step],yy[...,T+t:T+t+step],yy[...,2*T+t:2*T+t+step]],-1)
                #
                im = model(xx[...,:-1])
                #
                loss += myloss(im.reshape(batch_size, -1), y.reshape(batch_size, -1))

                if t == 0:
                    pred = im
                else:
                    pred = torch.cat((pred, im), -1)
                #
                xx_i = torch.cat([xx[..., step:T_in],               \
                                  torch.mean(im[...,0],1).reshape(batch_size,1,1).repeat(1,S,1),\
                                  xx[..., T_in+step:2*T_in],        \
                                  torch.mean(im[...,1],1).reshape(batch_size,1,1).repeat(1,S,1),\
                                  xx[...,2*T_in+step:3*T_in],       \
                                  im[...,2].reshape(batch_size,S,1) \
                                  ], dim=-1).to(device)
                #
                x_cl = torch.mean(im[...,0],1).reshape(batch_size,1,1).repeat(1,S,1)+im[...,2].reshape(batch_size,S,1)*np.cos(grid.repeat(batch_size,1,1)).to(device)
                y_cl = torch.mean(im[...,1],1).reshape(batch_size,1,1).repeat(1,S,1)+im[...,2].reshape(batch_size,S,1)*np.sin(grid.repeat(batch_size,1,1)).to(device)
                #
                k=0
                for i in xx[:,0,-1]:
                    ii = int(i.cpu().detach().numpy())
                    heter2[k,:,0] = my_to_torch(het(data_het[ii], \
                                                    x_cl[k,:,0].cpu().detach().numpy(), \
                                                    y_cl[k,:,0].cpu().detach().numpy()))   
                    k+=1
                #
                xx = torch.cat([xx_i,                               \
                                xx[...,3*T_in+step:4*T_in],         \
                                heter2.reshape(batch_size,S,1).to(device),     \
                                grid.repeat(batch_size,1,1).to(device),        \
                                xx[...,-1].reshape(batch_size,S,1), \
                                ], dim=-1    ) 

            test_l2_step += loss.item()
            test_l2_full += myloss(torch.cat([pred[...,0::3],pred[...,1::3],pred[...,2::3]],-1).reshape(batch_size, -1), yy.reshape(batch_size, -1)).item()

    t2 = default_timer()
    scheduler.step()
    #
    # Print the output
    #
    print('===========================')
    print('epoch = '+str(ep+1)+' (wall time = '+str(t2-t1)+')')
    print('===========================')
    print('training error (current step) = '+str(train_l2_step / ntrain / (T / step)))
    print('training error (total)        = '+str(train_l2_full / ntrain))
    print('testing error  (current step) = '+str(test_l2_step / ntest / (T / step)))
    print('testing error  (total)        = '+str(test_l2_full / ntest))
    #
    train_error = train_l2_full / ntrain
    test_error  = test_l2_full / ntest
    if test_error<min_error:
        print('Validation error reduced. Printing model parameters...')
        #torch.save(model, path_model+'_ep'+str(ep+1))
        #torch.save(model.state_dict(), path_model)
        torch.save({'epoch': ep, \
                    'model_state_dict': model.state_dict(), \
                    'optimizer_state_dict': optimizer.state_dict(), \
                    'scheduler_state_dict': scheduler.state_dict(), \
                    'loss': loss, \
                    }, path_model)
        min_error = test_error
    #
    # Plot the error
    #
    plot_error=np.concatenate([plot_error,[[train_error],[test_error]]],-1)
    plt.plot(np.linspace(0,ep+1,ep+1),plot_error[0,:],"r",label='training error')
    plt.plot(np.linspace(0,ep+1,ep+1),plot_error[1,:],"b",label='validation error')
    plt.legend()
    plt.yscale("log")
    plt.savefig(path_plots+'_errors.pdf',bbox_inches='tight',transparent=True)
    plt.clf()
    
    print(ep+1, train_error, test_error, file=open(path_model+'_err.out', 'a'))





